from typing import Any, Text, Dict, List

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.helpers.localization import _
from actions.helpers.conversation_tagging_feature import ConversationTagger
from actions.helpers.endpoint_diagnosis import EndpointDiagnosis

import logging

log = logging.getLogger(__name__)


class ActionHelloWorld(Action):

    def name(self) -> Text:
        return "action_hello_world"

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        # TODO This might become a helper "tracker.get_last_user_events(turn_count)"
        user_event_list = []
        for events in tracker.events_after_latest_restart():
            if events["event"] == "user":
                user_event_list.append(events)

        last_user_event = user_event_list[-1]

        metadata = last_user_event['metadata']

        log.info(f"Received metdata: {metadata}")
        log.info(f"Received user_id: {metadata.get('user_id')}")

        dispatcher.utter_message(text=_("Hello World"))
        dispatcher.utter_message(text=_("And again"))

        return []


class ActionRepeat(Action):
    '''
        Custom action repeating the last bot response.
        This action considers conditional responses and
        can handle the paraphrasing.

        Typical tracker list in Rasa 2.8.x: {
            'event': 'bot',
            'timestamp': 1637768897,
            'metadata': {'utter_action': 'utter_greet'},
            'text': 'Hey! How are you?',
            'data': {
                'elements': None,
                'quick_replies': None,
                'buttons': None,
                'attachment': None,
                'image': None,
                'custom': None}
        }
    '''

    def name(self) -> Text:
        return "action_repeat"

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        user_ignore_count = 2
        count = 0
        tracker_list: tracker.events = []

        '''
            Builds a bot event list.
            user_ignore_count indicates how many past turns are considered.
            count = 0 is the recent event and count = -1 is the event before.

            U: Hello.               (greet, Turn 1)
            B: Hello.               (response 1, T1)
            B. How can I help?      (response 2, T1, count = -4, ignore = 0)
            U: What can you do?     (capability, T2, count = -3, ignore = 1)
            B: ...                  (response 1, T2, count = -2, ignore = 1)
            B: ...                  (response 2, T2, count = -1, ignore = 1)
            U: What do you mean?    (paraphrase, T3, count = 0, ignore = 2)
            B: ...                  (repeat paraphrase, T3)
        '''
        # TODO We should consider having a helper to access turns quickly
        # TODO e.g., get_bot_events_from_turn(tracker, turn=1)
        while user_ignore_count > 0:
            event = tracker.events[count].get('event')
            if event == 'user':
                user_ignore_count = user_ignore_count - 1
            if event == 'bot':
                tracker_list.append(tracker.events[count])
            count = count - 1

        # Reverse the list to maintain the chronological order.
        tracker_list.reverse()

        for event in tracker_list:
            log.info(f"Tracker event: {event}")
            log.info(f"Meta data: {event.get('metadata')}")

            metadata: Dict[Text, Any] = event.get('metadata')

            # Differentiate the case where the bot utterance comes from:
            # 1. a response template dispatcher.utter_message(response='')
            #   1.1 conditional response
            #   1.2 regular response
            # 2. a plain text (dispatcher.utter_message(text='')).
            # ! Response templates are the recommended practice.
            if len(metadata) > 0:
                # Use text key, since this action doesn't recreate conditions.
                if 'condition' in metadata:
                    dispatcher.utter_message(text=event.get('text'))
                else:
                    dispatcher.utter_message(response=metadata.get('utter_action'))

            else:
                dispatcher.utter_message(text=event.get('text'))

        return []


class ActionActivateParaphrasing(Action):
    '''
        Sets the paraphrasing_activated slot to True.
        Used to repeat the paraphrased version of the latest responses.
    '''

    def name(self) -> Text:
        return "action_activate_paraphrasing"

    def run(self, dispatcher: CollectingDispatcher, tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        return [
            SlotSet('paraphrasing_activated', True),
        ]


class ActionDeactivateParaphrasing(Action):
    '''
        Sets the paraphrasing_activated slot to False.
        Used to avoid repeating the paraphrased version
        of the following respones.
    '''

    def name(self) -> Text:
        return "action_deactivate_paraphrasing"

    def run(self, dispatcher: CollectingDispatcher, tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        return [SlotSet('paraphrasing_activated', False)]


class ActionTagConversationPositive(Action):
    '''
        This action tags the current conversation.
        The tag_name slot provides the tag's name.
    '''

    def name(self) -> Text:
        return "action_tag_conversation_positive"

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        # TODO need to get tag name from a slot
        tag_name = "positive"
        conversation_id = tracker.sender_id

        log.info(f"Conversation for tagging: {conversation_id}")

        ConversationTagger().tag_conversation(conversation_id, tag_name)

        return []


class TagConversationNegative(Action):
    '''
        This action tags the current conversation.
        The tag_name slot provides the tag's name.
    '''

    def name(self) -> Text:
        return "action_tag_conversation_negative"

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        # TODO need to get tag name from a slot
        tag_name = "negative"
        conversation_id = tracker.sender_id

        log.info(f"Conversation for tagging: {conversation_id}")

        ConversationTagger().tag_conversation(conversation_id, tag_name)

        return []


class TagConversationForReview(Action):
    '''
        This action tags the current conversation.
        The tag_name slot provides the tag's name.
    '''

    def name(self) -> Text:
        return "action_tag_conversation_for_review"

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        tag_name = ""
        conversation_id = tracker.sender_id

        log.info(f"Conversation for tagging: {conversation_id}")

        ConversationTagger().tag_conversation(conversation_id, tag_name)

        return []


class DiagnoseEndpoints(Action):
    '''
        This action diagnoses the endpoints defined in the configs folder.
        The action dispatches one message per endpoint.
        It assumes that the debug_mode slot is set to True.
    '''

    def name(self) -> Text:
        return "action_diagnose_endpoints"

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        debug_mode = tracker.get_slot("debug_mode")

        if debug_mode:
            result = EndpointDiagnosis().run()

            for message in result:
                dispatcher.utter_message(text=message)

        return []
