# -*- coding: utf-8 -*-
import logging
import json

from typing import Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.types import DomainDict
from rasa_sdk.executor import CollectingDispatcher
from rasa_sdk.events import (
    UserUtteranceReverted,
    EventType,
)

# from actions.helpers.conversation_tagging_feature import ConversationTagger

USER_INTENT_OUT_OF_SCOPE = "out_of_scope"

log = logging.getLogger(__name__)

INTENT_DESCRIPTION_MAPPING_PATH = "actions/actions_cognitive_advisor/intent_description_mapping.csv"

class ActionDefaultAskAffirmation(Action):
    """Asks for an affirmation of the intent if NLU threshold is not met."""

    def name(self) -> Text:
        return "action_default_ask_affirmation"

    def __init__(self) -> None:
        import pandas as pd

        self.intent_mappings = pd.read_csv(INTENT_DESCRIPTION_MAPPING_PATH)
        self.intent_mappings.fillna("", inplace=True)
        self.intent_mappings.entities = self.intent_mappings.entities.map(
            lambda entities: {e.strip() for e in entities.split(",")}
        )

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: DomainDict,
    ) -> List[EventType]:


        print(" stage 1"
        )
        latest_message = tracker.latest_message
        if latest_message is None:
            return []
        print("stage 2")
        intent_ranking = latest_message.get("intent_ranking", [])
        print(intent_ranking)
        # log.info(f"Intent ranking is: {intent_ranking}")
        if len(intent_ranking) > 1:
            diff_intent_confidence = intent_ranking[0].get(
                "confidence"
            ) - intent_ranking[1].get("confidence")
            if diff_intent_confidence < 0.2:
                intent_ranking = intent_ranking[:2]
            else:
                intent_ranking = intent_ranking[:1]

        # for the intent name used to retrieve the button title, we either use
        # the name of the name of the "main" intent, or if it's an intent that triggers
        # the response selector, we use the full retrieval intent name so that we
        # can distinguish between the different sub intents
        first_intent_names = [
            intent.get("name", "") if intent.get("name", "") not in ["faq"]

            else tracker.latest_message.get("response_selector")
            .get(intent.get("name", ""))
            .get("ranking")[0]
            .get("intent_response_key")

            for intent in intent_ranking
        ]

        if "nlu_fallback" in first_intent_names:
            first_intent_names.remove("nlu_fallback")
        if "/out_of_scope" in first_intent_names:
            first_intent_names.remove("/out_of_scope")
        if "out_of_scope" in first_intent_names:
            first_intent_names.remove("out_of_scope")

        log.debug(f"intent ranking contains: {first_intent_names}")
        print("stage 3")
        print(first_intent_names)
        if len(first_intent_names) > 0:

            entities = tracker.latest_message.get("entities", [])
            entities = {e["entity"]: e["value"] for e in entities}

            entities_json = json.dumps(entities)

            buttons = []
            for intent in first_intent_names:
                button_title = self.get_button_title(intent, entities)
                if "/" in intent:
                    # here we use the button title as the payload as well, because you
                    # can't force a response selector sub intent, so we need NLU to parse
                    # that correctly
                    buttons.append({"title": button_title, "payload": button_title})
                else:
                    buttons.append(
                        {"title": button_title, "payload": f"/{intent}{entities_json}"}
                    )

            buttons.append({"title": "Iets anders", "payload": "/out_of_scope"})
            
            dispatcher.utter_message(response="utter_ask_to_verify_intent", buttons=buttons)
        else:

            dispatcher.utter_message(response="utter_ask_to_repeat")

        return []

    def get_button_title(self, intent: Text, entities: Dict[Text, Text]) -> Text:
        default_utterance_query = self.intent_mappings.intent == intent
        utterance_query = (self.intent_mappings.entities == entities.keys()) & (
            default_utterance_query
        )

        utterances = self.intent_mappings[utterance_query].button.tolist()

        if len(utterances) > 0:
            button_title = utterances[0]
        else:
            utterances = self.intent_mappings[default_utterance_query].button.tolist()
            button_title = utterances[0] if len(utterances) > 0 else intent

        return button_title.format(**entities)


class ActionDefaultFallback(Action):
    def name(self) -> Text:
        return "action_default_fallback"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: DomainDict,
    ) -> List[EventType]:

        # Fallback caused by TwoStageFallbackPolicy

        # returns the attributes of the latest message (here the intent name)
        last_intent = tracker.latest_message["intent"]["name"]
        if last_intent in ["nlu_fallback", USER_INTENT_OUT_OF_SCOPE]:
            # conversation_id = tracker.sender_id
            # ConversationTagger().tag_conversation(conversation_id, "Negative")
            # return [SlotSet("feedback_value", "negative")]
            dispatcher.utter_message(response="utter_default_two_stage_fallback")
            return [UserUtteranceReverted()]

        # Fallback caused by Core
        else:
            dispatcher.utter_message(response="utter_default")
            return [UserUtteranceReverted()]


class ActionRestartWithButton(Action):
    def name(self) -> Text:
        return "action_restart_with_button"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: DomainDict,
    ) -> None:

        dispatcher.utter_message(response="utter_restart_with_button")
