from typing import Any, Text, Dict, List

from rasa_sdk import Action, Tracker
from rasa_sdk.executor import CollectingDispatcher
from rasa_sdk.events import SlotSet, EventType
from rasa_sdk.types import DomainDict
from datetime import datetime, timedelta

from actions.helpers.WhyEngineClient import (
   WhyEngineClient,
   create_json_instance,
   parse_explanation
   )

from actions.helpers.PreciaDataQuery import (
   PreciaDataQuery
   )

import logging
logger = logging.getLogger(__name__)


class RequestSpeedRecommendation(Action):
    """
       Provide speed recommendation
    """
    def name(self) -> Text:
        return "action_request_speed_recommendation"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: DomainDict
    ) -> List[EventType]:
        '''
            Uses the tracker to get the required slots
        '''

        try:
            
            precia = PreciaDataQuery()

            # data = precia.query_precia_limit(500)
            

            data = precia.query_direct_precia(offset=0,limit=500)
            rows = len(data)
            logger.info("Rows queried from precia: "+str(rows))

            # function to extract relevant data to create the instance for the why engine
            speed,maxSpeed,avgSpeed,deltaMaxSpeed,latest_weight,density,cans,latest_weighingdate,productcode = precia.prepare_precia_data_why_engine(data)
            logger.info("Current speed: "+str(speed))

            latest_batch = data['batchinformation'].iloc[0]  # get the value of the latest batch
            # calculating batch_bum by finding when the last batch change occured and counting rows
            cum_count = data.groupby('batchinformation').cumcount()  # compute cum_count for each batch
            batch_num = cum_count[data['batchinformation'] == latest_batch].max() + 1  # count rows for latest batch

            # Parse the string into a datetime object
            dt = datetime.strptime(str(latest_weighingdate), "%Y-%m-%d %H:%M:%S")

            # Format the datetime object into the desired string format
            latest_weighingdate = dt.strftime("%H:%M:%S %d-%m")
            
            # latest_weighingdate = "N/A"
            # productcode = "N/A"
            # create instance for why engine
            # instance = create_json_instance(speed=0.5,WeightNet=5000,density=1.02,batch_num=50,maxSpeed=1,avgSpeed=0.5,deltaMaxSpeed=0.5,model_id='rf_2023_03_04_15_05_31',explanations=5)
            instance = create_json_instance(speed,latest_weight,density,batch_num,maxSpeed,avgSpeed,deltaMaxSpeed,model_id='rf_2023_03_04_15_05_31',explanations=5)
            logger.info("instance created")

            # create why engine client instance
            client = WhyEngineClient()

            # send instance to why engine
            why_engine_response = client.explain_instance(instance)

            classification,extracted_explanation = parse_explanation(why_engine_response)

            if classification == 0:
                # reduce
                advice = "verlagen"
            elif classification == 1:
                # no change
                advice = "zo houden"
            else:
                # increase
                advice = "verhogen"
            
            dispatcher.utter_message(f"Snelheid {advice} op basis van ...")

            # Access the first explanation
            first_features_list = extracted_explanation[0]
            for feature in first_features_list:
                feature_name = feature['feature_name']
                feature_value = feature['feature_value']
                if feature_name == 'speed' or feature_name == 'avgSpeed' or feature_name == 'deltaMaxSpeed':
                    feature_value = round(float(feature_value)*60,1)
                
                if feature_name == 'speed':
                    feature_name = 'Productiesnelheid'
                elif feature_name == 'avgSpeed':
                    feature_name = f"Gemiddelde productiesnelheid van SKU {productcode}"
                elif feature_name == 'deltaMaxSpeed':
                    feature_name = f"Delta tot de max productiesnelheid van SKU {productcode}"
                elif feature_name == 'batch_num':
                    feature_name = "Flakon nummer"
                elif feature_name == 'density':
                    feature_name = "Dichtheid"
                elif feature_name == "WeightNet":
                    feature_name = "Nettogewicht"
            
                dispatcher.utter_message(f"{feature_name} is {feature_value}")
            dispatcher.utter_message(f"Geupdate om {latest_weighingdate}")

        except Exception as e:
            logger.info("An error occurred: "+str(e))
            dispatcher.utter_message(
                    response="utter_no_data_timeframe",
                )

        # clear slot
        return []
