from typing import Any, Text
from graphqlclient import GraphQLClient
from urllib.error import URLError

import actions.helpers.config_loader as config
import logging

log = logging.getLogger("GraphQL queries")

# The Docker container information for the GraphQL server.
# Container name should be set with Docker's --name="graphql-server" parameter.
container_name = config.get_endpoint("graphql").get("container_name", "")
container_port = str(config.get_endpoint("graphql").get("container_port", ""))
routing_parameter = config.get_endpoint("graphql").get("routing_parameter", "")

graphql_url = f"http://{container_name}:{container_port}{routing_parameter}"
# TODO Add query to get possible causes, solutions and symptoms for an event.
# TODO Add query to add events, causes, solutions etc
# TODO Add query to add SKU settings (by creating a node with relationships to SKU, user, filler)

# TODO other filters for searchdoc?
# graphql query to find nodes by status
def graphql_findbystat(status: Text) -> Any:
    query = ('''query ($status: String!)
    {
                  findbystat(status: $status) {
                      type
                      name
                      date
                      }
                }
                ''')
    
    variables = {
        "status": status
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to change status of product settings
def graphql_createSkuStat(name: Text,
        status: Text,
        ) -> Any:
    
    query = ('''mutation ($name: String!, $status: String!)
    {
                    CreateSkuStat(skuStatData: {name: $name, status: $status}){
                        skuStat{
                            name
                            status
                        }
                    }
                }        
                ''') 
    
    variables = {
        "name": name,
        "status": status
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to change status of doc
def graphql_createDocStat(docDescriptionEn: Text,
        status: Text,
        ) -> Any:
    
    query = ('''mutation ($docDescriptionEn: String!, $status: String!)
    {
                    CreateDocStat(docStatData: {docDescriptionEn: $docDescriptionEn, status: $status}){
                        docStat{
                            docDescriptionEn
                            status
                        }
                    }
                }        
                ''') 
    
    variables = {
        "docDescriptionEn": docDescriptionEn,
        "status": status
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to add new rating for product settings
def graphql_createSkuRate(name: Text,
        rate: float,
        ) -> Any:
    
    query = ('''mutation ($name: String!, $rate: Float!)
    {
                    CreateSkuRate(skuRateData: {name: $name, rate: $rate}){
                        skuRate{
                            name
                            rate
                        }
                    }
                }        
                ''') 
    
    variables = {
        "name": name,
        "rate": rate
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to add new rating for documents
def graphql_createDocRate(docDescriptionEn: Text,
        rate: float,
        ) -> Any:
    
    query = ('''mutation ($docDescriptionEn: String!, $rate: Float!)
    {
                    CreateDocRate(docRateData: {docDescriptionEn: $docDescriptionEn, rate: $rate}){
                        docRate{
                            docDescriptionEn
                            rate
                        }
                    }
                }        
                ''') 
    
    variables = {
        "docDescriptionEn": docDescriptionEn,
        "rate": rate
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result
    
# Mutation to add new status for solution events
def graphql_createStatSol(descEn: Text,
        status: Text,
        ) -> Any:
    
    query = ('''mutation ($descEn: String!, $status: String!)
    {
                    CreateStatSol(statSolData: {descEn: $descEn, status: $status}){
                        statSol{
                            descEn
                            status
                        }
                    }
                }        
                ''') 
    
    variables = {
        "descEn": descEn,
        "status": status
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to add new rating for solution events
def graphql_createRateSol(descEn: Text,
        rate: float
        ) -> Any:
    
    query = ('''mutation ($descEn: String!, $rate: Float!)
    {
                    CreateRateSol(rateSolData: {descEn: $descEn, rate: $rate}){
                        rateSol{
                            EventDescriptionEnglish : descEn
                            rate : rate
                        }
                    }
                }        
                ''') 
    
    variables = {
        "descEn": descEn,
        "rate": rate
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result



# Mutation to create a relationship between a machine component and event
def graphql_createEvMcConn(descEn: Text,
        descNl: Text,
        descEn2: Text,
        descNl2: Text,
        name: Text,
        ) -> Any:
    
    query = ('''mutation ($descEn: String!,$descNl: String!,$descEn2: String!,$descNl2: String!,$name: String!)
    {
                    createEvMcConn(evMcConnData: {descEn: $descEn, descNl: $descNl, descEn2: $descEn2, descNl2: $descNl2, name: $name}){
                        evMcConn{
                            EventDescriptionEnglish : descEn
                            MachineComponentDescriptionEnglish : descEn2
                        }
                    }
                }        
                ''') 
    
    variables = {
        "descEn": descEn,
        "descNl": descNl,
        "descEn2": descEn2,
        "descNl2": descNl2,
        "name": name
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to create a cause relationship between events
def graphql_createCauseConn(descEn: Text,
        descEn2: Text) -> Any:
    
    query = ('''mutation ($descEn: String!,$descEn2: String!)
    {
                    createCauseConn(causeConnData: {descEn: $descEn, descEn2: $descEn2}){
                        causeConn{
                            event : descEn
                            cause : descEn2
                        }
                    }
                }        
                ''') 
    
    variables = {
        "descEn": descEn,
        "descEn2": descEn2
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to create a solution relationship between events
def graphql_createSolutionConn(descEn: Text,
        descEn2: Text) -> Any:
    
    query = ('''mutation ($descEn: String!,$descEn2: String!)
    {
                    createSolutionConn(solutionConnData: {descEn: $descEn, descEn2: $descEn2}){
                        solutionConn{
                            descEn : descEn
                            descEn2 : descEn2
                        }
                    }
                }        
                ''')
    
    variables = {
        "descEn": descEn,
        "descEn2": descEn2
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to create an error code node
def graphql_createErrCode(name: Text,
        descEn: Text) -> Any:
    
    query = ('''mutation ($name: String!,$descEn: String!)
    {
                    createErrCode(errCodeData: {name: $name, descEn: $descEn}){
                        errCode{
                            name : name
                            desc_en : descEn
                        }
                    }
                }        
                ''')
    
    variables = {
        "name": name,
        "descEn": descEn
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# Mutation to create an event node
def graphql_createEvent(descEn: Text,
        descNl: Text) -> Any:
    
    query = ('''mutation ($descEn: String!,$descNl: String!)
    {
                    createEvent(newEventData: {descEn: $descEn, descNl: $descNl}){
                        newEvent{
                            descEn : descEn
                            descNl : descNl
                        }
                    }
                }        
                ''')
    
    variables = {
        "descEn": descEn,
        "descNl": descNl
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# graphql query to retrieve solutions
def graphql_findsolution(descEn: Text,name: Text, descEn3: Text) -> Any:
    query = ('''query ($descEn: String!,$name: String!,$descEn3: String!)
    {
                  findsolution(descEn: $descEn,name: $name, descEn3: $descEn3) {
                      descEn2
                      descNl2
                      avrate
                      status
                      }
                }
                ''')
    
    variables = {
        "name": name,
        "descEn": descEn,
        "descEn3": descEn3
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# graphql query to retrieve causes
def graphql_findcause(name: Text, datetime: Text, errName: Text, descEn: Text) -> Any:
    query = ('''query ($name: String!, $datetime: String!, $errName: String!, $descEn: String!)
    {
                  findcause(name: $name, datetime: $datetime, errName: $errName, descEn: $descEn) {
                      descriptionEnglish 
                      descriptionDutch 
                      }
                }
                ''')
    
    variables = {
        "name": name,
        "datetime": datetime,
        "errName": errName,
        "descEn": descEn
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result
    
def graphql_searchdoc(machineCompName: Text) -> Any:
    query = ('''query ($machineCompName: String!)
    {
                  searchdoc(machineCompName: $machineCompName) {
                      nameDoc
                      docDescriptionEn
                      docType
                      docUrlNl
                      docVersion
                      timeRequired
                      avrate
                      status
                      }
                }
                ''')
    
    variables = {
        "machineCompName": machineCompName
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result
# graphql query to retrieve SKU filler settings
def graphql_searchset(node1name: Text) -> Any:
    query = ('''query ($node1name: String!)
    {
                  searchset( node1name: $node1name) {
                      fillingPressure
                      fillingSpeed
                      comments
                      datetime
                      avrate
                      status
                      name
                      }
                }
                ''')
    
    variables = {
        "node1name": node1name
    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

def graphql_createSetting(name: Text,
        fillingPressure: Text,
        fillingSpeed: Text,
        comments: Text,
        dateTime: Text) -> Any:
    
    query = ('''mutation ($name: String!,$fillingSpeed: String!,$fillingPressure: String!,$comments: String!,$dateTime: String!)
    {
                    createSetting(settingData: {name: $name, fillingSpeed: $fillingSpeed, fillingPressure: $fillingPressure, comments: $comments, dateTime: $dateTime}){
                        setting{
                            settings_name: name
                            fillingSpeed
                            fillingPressure
                            dateTime
                        }
                    }
                }        
                ''')
    
    variables = {
        "name": name,
        "fillingSpeed": fillingSpeed,
        "fillingPressure": fillingPressure,
        "comments": comments,
        "dateTime": dateTime

    }
    
    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

def graphql_createConnectionSkuSet(
        sku_name: Text,
        Settings_name: Text,
        name: Text) -> Any:
    
    query = ('''mutation ($node1Name: String!,$node2Name: String!,$name: String!){
                    createConnectionSkuSet( connectionSkuSetData: {name: $name, label1:"Sku", node1Name: $node1Name, label2:"SKU_setting", node2Name: $node2Name}){
                        connectionSkuSet{
                            Connection_name : name
                            sku_name : node1Name
                            Settings_name : node2Name
                        }
                    }
                }        
                ''')

    variables = {
        "node1Name": sku_name,
        "node2Name": Settings_name,
        "name": name
    }

    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query,variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

def graphql_createConSetMc(
        Machine_component: Text,
        Settings_name: Text,
        Connection_name: Text
        ) -> Any:
    # TODO is it possible to generalise the code so we can create connections between any nodes?
    query = ('''mutation ($node1Name: String!,$node2Name: String!,$name: String!){
                    createConSetMc( conSetMcData: {name:$name, label1:"Machine_component", node1Name:$node1Name, label2:"SKU_setting", node2Name:$node2Name}){
                        conSetMc{
                            Connection_name : name
                            Machine_component_name : node1Name
                            Settings_name : node2Name
                        }
                    }
                }        
                ''')

    variables = {
        "node1Name": Machine_component,
        "node2Name": Settings_name,
        "name": Connection_name
    }

    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query,variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

def graphql_createSetConOp(
        user_id: Text,
        Settings_name: Text,
        connection_name: Text) -> Any:
    
    query = ('''mutation ($node1Name: String!,$node2Name: String!,$name: String!){
                    createSetConOp( setConOpData: {name:$name, label1:"Operator", node1Name:$node1Name, label2:"SKU_setting", node2Name:$node2Name}){
                        setConOp{
                            Connection_name : name
                            Operator_name : node1Name
                            Settings_name : node2Name
                        }
                    }
                }        
                ''')

    variables = {
        "node1Name": user_id,
        "node2Name": Settings_name,
        "name" : connection_name
    }

    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query,variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

def graphql_issue_handling_support(machine_component: Text,
        machine_component_state: Text,
        product_component: Text,
        product_component_state: Text,
        sku: Text,
        ) -> Any:
    
    # TODO need to write query for issue handling support
    query = ('''query {
                    SKU_filler_setting{
                      datetime
                      filling_pressure
                      user_id
                      filling_speed
                  }
                }''')

    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# TODO create a seperate query for symptom, cause and solution and use datetime to match them?
def graphql_report_issue(machine_component: Text,
        machine_component_state: Text,
        product_component: Text,
        product_component_state: Text,
        sku: Text,
        datetime: Text,
        ) -> Any:
    
    # TODO need to write query for recording issue report
    query = ('''query {
                    SKU_filler_setting{
                      datetime
                      filling_pressure
                      user_id
                      filling_speed
                  }
                }''')

    client = GraphQLClient(graphql_url)

    try:

        result = client.execute(query)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    return result

# TODO create query to create relation between user and document (when retrieved)

def graphql_create_user_SKU_settings_relation(
        user_id: Text,
        datetime: Text) -> Any:
    # TODO need to write query for creating user SKU settings relation

    # this query builds a relationship between a user profile and a product.
    '''
      Returns the results of the query from graphql server.

      The query structure in f-strings is this:
        f"{{ failures( {filters} ) {{ {attributes} }} }}"

      Example result:
      {
        "data": {
          "create_user_product_relation": "HAS_SCANNED"
        }
      }

    '''
    # Create a graphQL client object to query from GraphQL server
    client = GraphQLClient(graphql_url)

    query = ('''query ($user_id: String!, $product_id: String!)
    {
      create_user_product_relation (user_id:$user_id, product_id:$product_id)
    }''')

    variables = {
        "user_id": user_id,
        "datetime": datetime
    }

    # log.info("Query without filter for GraphQL server" +
    #         f"({graphql_url}): {query}, {variables}")

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    # TODO add other errors if needed

    return result

# queries beyond here are from BIBA - for reference.
def graphql_get_checklist_step_info(
        expertise_level: Text,
        checklist_id: Text,
        current_step: Text) -> Any:
    '''
      Returns the results of the query from graphql server.

      The query structure in f-strings is this:
        f"{{ failures( {filters} ) {{ {attributes} }} }}"

      Example result:
        {
          "data": {
            "users": [
              {
                "expertise_level": "Expert",
                "checklists": [
                  {
                    "checklist_id": "1",
                    "checklist_language": "en",
                    "first_step": "1",
                    "last_step": "13",
                    "steps": [
                      {
                        "step": "1",
                        "step_id": "1",
                        "next_step": [
                          {
                            "step": "2",
                            "step_id": "2"
                          }
                        ],
                        "previous_step": [],
                        "components": [
                          {
                            "component": "Instruction for Use with focus on labels",
                            "objects": [
                              {
                                "objective": "Right labels, good printing quality",
                                "actions": [
                                  {
                                    "action": "Inspect visually, use the info systems"
                                  }
                                ],
                                "defects": [
                                  {
                                    "defect": "Wrong labels, bad printing quality"
                                  }
                                ]
                              }
                            ]
                          }
                        ]
                      }
                    ]
                  }
                ]
              }
            ]
          }
        }
    '''

    checklist_language = config.get_language()

    # Create a graphQL client object to query from GraphQL server
    client = GraphQLClient(graphql_url)

    query = ('''query ($expertise_level: String!, $checklist_language: String!,
     $checklist_id: String!, $current_step: String!)
    {
      users (expertise_level: $expertise_level, checklist_language: $checklist_language,
       checklist_id: $checklist_id, current_step: $current_step){
        expertise_level
        checklists {
          checklist_id
          checklist_language
          first_step
          last_step
          steps {
            step
            step_id
            time
            next_step {
              step
              step_id
            }
            previous_step {
              step
              step_id
            }
            components {
              component
              objects {
                objective
                actions {
                  action
                }
                defects {
                  defect
                }
              }
            }
          }
        }
      }
    }''')

    variables = {
        "expertise_level": expertise_level,
        "checklist_language": checklist_language,
        "checklist_id": checklist_id,
        "current_step": current_step
    }

    # log.info("Query without filter for GraphQL server" +
    #         f"({graphql_url}): {query}, {variables}")

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    # TODO add other errors if needed

    return result


# This is the new query that we use to get checklist step info
def graphql_get_product_profile_active_checklist_step(
        user_id: Text,
        product_id: Text) -> Any:
    '''
      Returns the results of the query from graphql server.

      The query structure in f-strings is this:
        f"{{ failures( {filters} ) {{ {attributes} }} }}"

      Example result:
      {
        "data": {
          "get_product_profile_active_checklist_step": [
            {
              "expertise_level": "Expert",
              "checklists": [
                {
                  "checklist_id": "1",
                  "checklist_language": "en",
                  "first_step": "1",
                  "last_step": "13",
                  "steps": [
                    {
                      "step": "10",
                      "step_id": "10",
                      "time": 20,
                      "next_step": [
                        {
                          "step": "11",
                          "step_id": "11",
                          "time": 30
                        }
                      ],
                      "previous_step": [
                        {
                          "step": "9",
                          "step_id": "9"
                        }
                      ],
                      "components": [
                        {
                          "component": "External covers",
                          "objects": [
                            {
                              "objective": "External covers do not have aesthetical
                               issues and are well assembled",
                              "actions": [
                                {
                                  "action": "Inspect visually according to aesthetical catalog
                                   definitions and check correspondence in COALA tablet
                                   in case of doubt"
                                }
                              ],
                              "defects": [
                                {
                                  "defect": "Scratch, dents, painting mismatch, dots,
                                   missing parts, bad assembly"
                                }
                              ]
                            }
                          ]
                        }
                      ]
                    }
                  ]
                }
              ]
            }
          ]
        }
      }
    '''

    # Create a graphQL client object to query from GraphQL server
    client = GraphQLClient(graphql_url)

    query = ('''query ($user_id: String!, $product_id: String!){
    get_product_profile_active_checklist_step (user_id:$user_id, product_id:$product_id) {
      expertise_level
      checklists {
        checklist_id
        checklist_language
        first_step
        last_step
        steps {
          step
          step_id
          time
          next_step {
            step
            step_id
            time
          }
          previous_step {
            step
            step_id
          }
          components {
            component
            objects {
              objective
              actions {
                action
              }
              defects {
                defect
              }
            }
          }
        }
      }
    }
    }''')

    variables = {
        "user_id": user_id,
        "product_id": product_id
    }

    # log.info("Query without filter for GraphQL server" +
    #         f"({graphql_url}): {query}, {variables}")

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    # TODO add other errors if needed

    return result


def graphql_get_user_profile(
        user_id: Text) -> Any:
    '''
      Returns the results of the query from graphql server.

      The query structure in f-strings is this:
        f"{{ failures( {filters} ) {{ {attributes} }} }}"

      Example result:
      {
        "data": {
          "user_profile": [
            {
              "experience_level": "2",
              "user_id": "345abcd",
              "user_language": "en"
            }
          ]
        }
      }

    '''
    # Create a graphQL client object to query from GraphQL server
    client = GraphQLClient(graphql_url)

    query = ('''query ($user_id: String!)
    {
      user_profile (user_id:$user_id){
        experience_level
        user_id
        user_language
      }
    }''')

    variables = {
        "user_id": user_id
    }

    # log.info("Query without filter for GraphQL server" +
    #         f"({graphql_url}): {query}, {variables}")

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    # TODO add other errors if needed

    return result


def graphql_create_user_product_relation(
        user_id: Text,
        product_id: Text) -> Any:

    # this query builds a relationship between a user profile and a product.
    '''
      Returns the results of the query from graphql server.

      The query structure in f-strings is this:
        f"{{ failures( {filters} ) {{ {attributes} }} }}"

      Example result:
      {
        "data": {
          "create_user_product_relation": "HAS_SCANNED"
        }
      }

    '''
    # Create a graphQL client object to query from GraphQL server
    client = GraphQLClient(graphql_url)

    query = ('''query ($user_id: String!, $product_id: String!)
    {
      create_user_product_relation (user_id:$user_id, product_id:$product_id)
    }''')

    variables = {
        "user_id": user_id,
        "product_id": product_id
    }

    # log.info("Query without filter for GraphQL server" +
    #         f"({graphql_url}): {query}, {variables}")

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    # TODO add other errors if needed

    return result


def graphql_get_product_profile(
        product_id: Text) -> Any:
    # This query retrieves a dic of all valid products in the database

    '''
      Returns the results of the query from graphql server.

      The query structure in f-strings is this:
        f"{{ failures( {filters} ) {{ {attributes} }} }}"

      Example result when the product exists:
      {
        "data": {
          "product_profile": [
            {
              "product_identifier": "4898643HKJGD",
              "product_identifier_type": "SKU",
              "checklist_id": "1",
              "checklist_is_active": false,
              "active_checklist_step": [
                "1"
              ],
              "scanned_users": [
                {
                  "user_expertise_level": "Expert",
                  "user_id": "12345abcd",
                  "user_language": "en"
                }
              ],
              "status": [
                {
                  "name": "registered"
                }
              ]
            }
          ]
        }
      }
    '''
    # Create a graphQL client object to query from GraphQL server
    client = GraphQLClient(graphql_url)

    query = ('''query ($product_id: String!){
      product_profile (product_id:$product_id){
      product_identifier
      product_identifier_type
      checklist_id
      checklist_is_active
      active_checklist_step
      checklist_is_completed
      scanned_users {
        user_expertise_level
        user_id
        user_language
      }
      status {
        name
       }
      }
    }''')

    variables = {
        "product_id": product_id
    }

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    # TODO add other errors if needed

    return result


def graphql_update_product_profile(
        product_id: Text,
        checklist_is_active: bool,
        active_checklist_step: Text,
        checklist_is_completed: bool) -> Any:

    # this query write the checklist status and current step in the
    # product profile.
    '''
      Returns the results of the query from graphql server.

      The query structure in f-strings is this:
        f"{{ failures( {filters} ) {{ {attributes} }} }}"

      Example result:
      {"data": {
          "update_product_profile": [
            {
              "product_identifier": "4898643HKJGD",
              "product_identifier_type": "SKU",
              "checklist_id": "1",
              "checklist_is_active": true,
              "active_checklist_step": [
                "1"
              ],
              "scanned_users": [
                {
                  "user_expertise_level": "Expert",
                  "user_id": "12345abcd",
                  "user_language": "en"
                }
              ],
              "status": [
                {
                  "name": "registered"
                }
              ]
            }
          ]
        }
      }
    '''
    # Create a graphQL client object to query from GraphQL server
    client = GraphQLClient(graphql_url)

    query = ('''query ($product_id: String!, $checklist_is_active: Boolean!,
     $active_checklist_step: String!, $checklist_is_completed: Boolean!)
    {
      update_product_profile (product_id:$product_id,
        checklist_is_active:$checklist_is_active,
        active_checklist_step:$active_checklist_step,
        checklist_is_completed:$checklist_is_completed) {
        product_identifier
        product_identifier_type
        checklist_id
        checklist_is_active
        active_checklist_step
        scanned_users {
          user_expertise_level
          user_id
          user_language
        }
        status {
          name
        }
      }
    }''')

    variables = {
        "product_id": product_id,
        "checklist_is_active": checklist_is_active,
        "active_checklist_step": active_checklist_step,
        "checklist_is_completed": checklist_is_completed
    }

    try:

        result = client.execute(query, variables)

    except URLError as e:

        log.warn('Error while executing GraphQL query.')
        log.warn(f"Reason: {e.reason}")
        log.warn(f"Endpoint: {client.endpoint}")

        result = {
            "errors": [
                {
                    "error_response": "url_error",
                    "service_name": "GraphQL"
                }
            ]
        }

    # TODO add other errors if needed

    return result
