
import json
import requests
from datetime import datetime
import pandas as pd

import actions.helpers.config_loader as config
import numpy as np
from actions.helpers.UBI_token_manager import TokenManager
import actions.helpers.config_loader as config

import logging
logger = logging.getLogger(__name__)

class PreciaDataQuery:
    def __init__(self):
        self.url = config.get_endpoint("graphql_query").get("url", "")
        self.precia_url = config.get_endpoint("precia").get("url", "")
        self.token_manager = TokenManager()

    def query_direct_precia(self,offset=0,limit=500):
        precia_url = f"{self.precia_url}?offset={offset}&limit={limit}"
        response = requests.get(precia_url)
        if response.status_code == requests.codes.ok:
            # Convert JSON string to Python object
            json_data = json.loads(response.content)
            # data_array = json_data['data']
            # convert array to pandas DataFrame
            data = pd.json_normalize(json_data)
            # make all headers lower case to match mongodb/graphql
            data.columns = map(str.lower, data.columns)
            return data
        else:
            # Log useful information from the response
            logger.info("Request failed with status code: %d", response.status_code)
            logger.info("Response headers: %s", response.headers)
            logger.info("Response content: %s", response.content)
            raise Exception("Query failed to run by returning code of {}.".format(response.status_code))

    def query_precia_time_intervals(self,startdatetime,enddatetime):
        query = """
            query ListPreciaDataSpecificTimestampInterval_resolver($startdate: String!, $enddate: String!) {
            listPreciaDataSpecificTimestampInterval_resolver(startdate: $startdate, enddate: $enddate) {
                id
                createddate
                index
                i_d
                source
                hash
                weighingdate
                weightnet
                zonecode
                status
                density
                productcode
                batchinformation
                batchstartdate
            }
            }
        """

        headers = {
            "Content-Type": "application/json",
            "Authorization": self.token_manager.get_bearer_token()
        }

        data = {
            "query": query,
            "variables": {
                "startdate": startdatetime,
                "enddate": enddatetime
            }
        }

        response = requests.post(self.url, headers=headers, json=data)
        if response.status_code == 200:
            # Convert JSON string to Python object
            json_data = json.loads(response.content)
            data_array = json_data['data']['listPreciaDataSpecificTimestampInterval_resolver']
            # convert array to pandas DataFrame
            data = pd.json_normalize(data_array)
            return data
        else:
            raise Exception("Query failed to run by returning code of {}. {}".format(response.status_code, query))
        
    def query_precia_limit(self,limit):
        query = """
            query ListPreciaDatawithLimit($limit: Int!) {
            listPreciaDatawithLimit(limit: $limit) {
                id
                createddate
                index
                i_d
                source
                hash
                weighingdate
                weightnet
                zonecode
                status
                density
                productcode
                batchinformation
                batchstartdate
                }
            }
        """

        headers = {
            "Content-Type": "application/json",
            "Authorization": self.token_manager.get_bearer_token()
        }

        data = {
            "query": query,
            "variables": {
                "limit": limit
            }
        }

        response = requests.post(self.url, headers=headers, json=data)
        if response.status_code == 200:
            print(response.content)
            # Convert JSON string to Python object
            json_data = json.loads(response.content)
            data_array = json_data['data']['listPreciaDatawithLimit']
            # convert array to pandas DataFrame
            data = pd.json_normalize(data_array)
            return data
        else:
            raise Exception("Query failed to run by returning code of {}. {}".format(response.status_code, query)) 
        
    def query_precia_product(self):
        query = """
            query ListPreciaDatawithLimit($limit: Int!) {
            listPreciaDatawithLimit(limit: $limit) {
                id
                createddate
                index
                i_d
                source
                hash
                weighingdate
                weightnet
                zonecode
                status
                density
                productcode
                batchinformation
                batchstartdate
            }
            }

        """

        headers = {
            "Content-Type": "application/json",
            "Authorization": self.token_manager.get_bearer_token()
        }

        data = {
            "query": query,
            "variables": {
                "limit": 1
            }
        }

        response = requests.post(self.url, headers=headers, json=data)
        if response.status_code == 200:
            # Convert JSON string to Python object
            json_data = json.loads(response.content)
            data_array = json_data['data']['ListPreciaDatawithLimit']
            # convert array to pandas DataFrame
            data = pd.json_normalize(data_array)
            productcode = data.iloc[0]['productcode']

            return productcode
        else:
            raise Exception("Query failed to run by returning code of {}. {}".format(response.status_code, query))
        
    # To generate speed feature used to construct feature vectors
    def generate_speed(self, data, rows=1):
        # Sort data by weighingdate in descending order
        data = data.sort_values(by='weighingdate', ascending=False)
        tmp = list(data['weighingdate'])
        # calculate time delta to previous row
        tmp_diff = [abs((tmp[i] - tmp[i-1])).total_seconds() if i != 0 else 0 for i in range(len(tmp))]
        # calculate speed
        tmp_speed = [0 if diff == 0 else rows / diff for diff in tmp_diff]
        # round to 2 decimal places
        tmp_speed = [round(speed, 2) for speed in tmp_speed]

        return tmp_speed,tmp_diff
    
    def generate_latest_speed(self, data, rows=1):
        # Sort data by weighingdate in descending order
        data = data.sort_values(by='weighingdate', ascending=False)
        # Select the last X rows
        data_last_x_rows = data.iloc[:rows]
        
        # Get the weighingdate values of the last X rows
        tmp = list(data_last_x_rows['weighingdate'])
        
        # calculate time delta to previous row
        diff = abs((tmp[0] - tmp[rows-1]).total_seconds())
        
        # calculate speed and round
        speed = round(diff / rows,2)
        
        return speed

    def select_past_n_hours(self,data,hours):
        # Get current timestamp
        current_time = pd.Timestamp.now()

        # Subtract n hours from current timestamp
        time_threshold = current_time - pd.Timedelta(hours=hours)

        # Select rows with weighingdate within last n hours
        recent_data = data[data['weighingdate'] > time_threshold]

        return recent_data
    
    def clean_data(self,data):
        # this index check is not needed when using graphql/mongodb data
        # if "NumberWeighing" not in data.columns:
        #     data = data.reset_index().rename(columns={"index": "NumberWeighing"})
        #     data["NumberWeighing"] = data["NumberWeighing"] + 1
        # select good rows
        data = data[data['weightnet'] != 0]
        data = data.dropna(subset=['weightnet'])
        data = data.reset_index(drop=True)
        data['weighingdate'] = data['weighingdate'].apply(lambda x: datetime.strptime(x, '%Y-%m-%d  %H:%M:%S'))

        return data
    
    # a function to look up the max and avg previous speeds for a given product code
    def get_max_avg_speed(self,productcode, max_avg_speed_df):
        # Look up the corresponding row in the "max_avg_speed.csv" DataFrame
        row = max_avg_speed_df[max_avg_speed_df['ProductCode'] == productcode]
        
        # todo: calculate averages or go for something higher?
        # Check if the row is empty and assign default max and average speeds
        if row.empty:
            max_speed = 1.5
            avg_speed = 0.7
        else:
            # Extract the max and avg previous speeds for the product code
            max_speed = row['maxSpeed'].values[0]
            avg_speed = row['avgSpeed'].values[0]
            
        return max_speed, avg_speed

    def prepare_precia_data_why_engine(self,data):

        data = self.clean_data(data)

        cans = len(data)

        if cans > 120:
            rows = 120
        else:
            rows = cans

        speed = self.generate_latest_speed(data,rows)/60

        data_sorted = data.sort_values(by='weighingdate', ascending=False)
        latest_weight = data_sorted.iloc[0]['weightnet']
        latest_weighingdate = data_sorted.iloc[0]['weighingdate']
        density = data_sorted.iloc[0]['density']
        batch_num = cans
        productcode = data_sorted.iloc[0]['productcode']

        df_max_avg_speed = pd.read_csv('actions/helpers/why_engine/max_avg_speed.csv')

        maxSpeed, avgSpeed = self.get_max_avg_speed(productcode,df_max_avg_speed)

        deltaMaxSpeed = maxSpeed - speed

        return speed,maxSpeed,avgSpeed,deltaMaxSpeed,latest_weight,density,cans,latest_weighingdate,productcode


    def process_precia_data(self,data,hours):

        data = self.clean_data(data)

        # todo: remove comment after debug
        # not needed when using time interval query
        # data = self.select_past_n_hours(data,hours)

        # create list of unique batch numbers, products, and number of cans
        batches = data['batchinformation'].unique()
        products = data['productcode'].unique()
        cans = len(data)

        speed,delta = self.generate_speed(data,rows=1)
        data['speed'] = speed
        # set speeds below 0.2 to zero and calculate rolling average for speeds (excluding zeros)
        data['filtered_speed'] = data['speed'].apply(lambda x: x if x >= 0.2 else 0)
        data['rolling_speed_avg'] = data['filtered_speed'].replace(0, np.nan).rolling(window=30, min_periods=1).mean()

        # replace speed column with filtered_speed
        data['speed'] = data['filtered_speed']
        data['delta'] = delta
        # Select all rows where 'delta' is over 5 seconds (to calculate downtime)
        downtime_rows_delta = data[data['delta'] > 7]
        uptime_rows_delta = data[data['delta'] <= 7]
        # uptime_rows = data[data['delta'] < 5]
        downtime_under_180 = data[(data['delta'] >= 7) & (data['delta'] <= 180)]
        # downtime_over_180 = data[data['delta'] > 180]

        # number of stops under 3 minutes
        num_downtime_under_180 = downtime_under_180.shape[0]
        downtime_under_180_series = pd.Series(downtime_under_180['delta'])

        # total uptime
        uptime_delta_series = pd.Series(uptime_rows_delta['delta'])
        uptime = int(abs(uptime_delta_series.sum()))

        # total duration of stops under 3 minutes
        downtime_under_180 = round((abs(downtime_under_180_series.sum())/3600),1)

        # number of stops over 3 minute (this will also count changeovers which is why we subject the number of batches)
        # num_downtime_over_180 = downtime_over_180.shape[0]-batches+1

        # Sum all the values of 'delta' in the selected rows
        delta_series = pd.Series(downtime_rows_delta['delta'])
        total_downtime = int(abs(delta_series.sum()))

        # convert hours to seconds
        seconds = hours*3600
        # uptime = seconds-total_downtime
        availability = int((uptime/seconds)*100)

        # TODO need to check calculations. it seems like uptime is through this off (calculating too high) so i multiple by 50 instead of 60.
        average_speed = int(round(cans / uptime,3)*50)

        # split the data by Batch numbers
        dfs = [group[1] for group in data.groupby('batchinformation')]
        # get the changeover duration for each batch (by looking at the first and last two deltas)
        deltas = [df['delta'].iloc[0:2].sum() + df['delta'].tail(2).sum() for df in dfs]
        delta_series_changeover = pd.Series(deltas)
        changeover_time = int(abs(delta_series_changeover.sum()))
 
        return [data,batches,products,cans,total_downtime,availability,average_speed,changeover_time,num_downtime_under_180,downtime_under_180,uptime]