from typing import Text
import requests

import actions.helpers.config_loader as config

import logging

log = logging.getLogger("Conversation Tagger")


class ConversationTagger:
    '''
        This class manages the tagging process for Rasa
        conversations. Credentials to Rasa X are in the api crendentials file.
        The key names are a compromise between CI/CD variables (underscores)
        and the key names expected by the Rasa X REST API (no underscores).
    '''

    def __init__(self):
        self.container_name = config.get_endpoint("rasa-x").get("container_name", "")
        self.container_port = config.get_endpoint("rasa-x").get("container_port", "")
        self.username = config.get_credentials("rasa-x").get("user_name", "")
        self.password = config.get_credentials("rasa-x").get("password", "")
        self.token = self.get_rasa_x_token()

    def get_rasa_x_token(self) -> Text:
        '''Uses credentials to request a Rasa X access token.'''

        payload = {'username': self.username, 'password': self.password}

        # TODO need to handle more exceptions (HTML Errors and connection errors)
        try:
            result = requests.post(
                url=f"http://{self.container_name}:{self.container_port}/api/auth",
                json=payload)
        except ConnectionError as e:
            log.error(e)
            return ""

        return result.json().get("access_token", {})

    def tag_conversation(self, conversation_id: Text, tag_name: Text) -> None:
        '''
            Assigns a tag to the Rasa X conversation.
        '''
        url = f"http://{self.container_name}:{self.container_port}/api/conversations/{conversation_id}/tags"
        headers = {'Authorization': f"Bearer {self.token}"}

        # Map short name to payload.
        # value: the tag's name that will appear in Rasa X.
        # color: the hex color code without # for the tag.
        # id: ID 1 is the tag for "needs review".
        if tag_name == "negative":
            payload = [{"value": "negative feedback", "color": "FF0000"}]
        elif tag_name == "positive":
            payload = [{"value": "positive feedback", "color": "00FF00"}]
        else:
            payload = [{"id": 1}]

        # TODO need to handle more exceptions (HTML Errors and connection errors)
        try:
            result = requests.post(url=url, headers=headers, json=payload)
        except ConnectionError as e:
            log.error(e)
        else:
            # Warn the admin about HTML status codes.
            if result.status_code == 404:
                log.warn("Page not found. Check the " +
                         "host and conversation ID for Rasa X.")
