from typing import List, Text

import requests
from requests.exceptions import (
    ConnectTimeout,
    HTTPError,
    ReadTimeout,
    Timeout,
    ConnectionError
)
from urllib3.connection import HTTPConnection
import json
import actions.helpers.config_loader as config

import logging
log = logging.getLogger("Endpoint Diagnosis")


class EndpointDiagnosis:
    '''
        This class runs a dignosis of the endpoints defined
        in the endpoints.yaml file in the config folder.
        The diagnosis returns one string summarizing the result.

        Diagnosis checks for connection errors and 400+ HTML status codes.
    '''

    def __init__(self):
        pass

    def run(self) -> List[Text]:
        '''
            Checks the endpoints and returns the summary text.
            It calls the url defined with the health key to test
            if the response is OK (less than 400).
            The call assumes a GET query and
            that an access token might be passed.
        '''

        # TODO Load endpoint names from file
        endpoint_names = ["graphql", "rasa-x", "gitlab"]

        status_list = []

        for endpoint_name in endpoint_names:

            url = config.get_endpoint(endpoint_name).get("health")
            token = config.get_credentials("gitlab").get("token", "")
            headers = {'PRIVATE-TOKEN': token}

            # First part catches errors.
            try:
                response = requests.get(url=url, headers=headers)
            except (
                ConnectTimeout,
                HTTPError,
                ReadTimeout,
                Timeout,
                ConnectionError
            ) as e:
                log.error(e)
                status_list.append(
                    f"DEBUG: Connection problem with {endpoint_name}. Check the logs.")
            except HTTPConnection as connection:
                log.error(connection)
                status_list.append(
                    f"DEBUG: Connection problem with {endpoint_name}. Check the logs.")
            else:

                # Second part handles response codes 400+ (not ok).
                if response.ok:
                    status_list.append(f"OK for {endpoint_name}. ")
                else:
                    # TODO Check if the error message is always present. If so,
                    # TODO remove the default message.
                    errors = json.loads(response.content).get(
                        'errors', ['{"message": "No error message."}']
                    )

                    error_messages = ""

                    for error in errors:
                        error_messages = error_messages.join(error.get('message'))

                    status_list.append(
                        f"DEBUG: Problem {response.status_code} with " +
                        f"{endpoint_name}. {error_messages}"
                    )

        return status_list
