import requests
import json
from datetime import datetime
import actions.helpers.config_loader as config
import pytz

class Neo4jQuery:
    def __init__(self):
        self.auth = config.get_credentials("neo4j").get("auth", "")
        self.url = config.get_endpoint("neo4j").get("url", "")
        self.headers = {
                'Content-Type': 'application/json',
                'Authorization': self.auth
            }
        
    def get_solutions(self, error):
        payload = {
            "statements": [
                {
                    "statement": """
                        MATCH (solution:Event)-[:SOLUTION_OF]->(event:Event)-[:HAS_CODE]->(error_code:Error_code)
                        WHERE error_code.name = $name
                        RETURN solution
                    """,
                    "parameters": {
                        "name": error
                    },
                }
            ]
        }
        try:
            response = requests.post(self.url, headers=self.headers, data=json.dumps(payload))
            result = response.json()
        except requests.exceptions.RequestException as e:
            return "error", [str(e)]

        solutions_list = []

        if "errors" in result and len(result["errors"]) > 0:
            error_messages = [error["message"] for error in result["errors"]]
            return "error", error_messages
        else:
            solutions = result["results"][0]["data"]
            if len(solutions) > 0:
                for i, solution_data in enumerate(solutions):
                    solution_attributes = solution_data["row"][0]

                    datetime_value = int(solution_attributes['datetime'])
                    desc_nl_value = solution_attributes['desc_nl']

                    # Convert the timestamp to a human-readable date and time format
                    datetime_object = datetime.fromtimestamp(datetime_value)
                    formatted_datetime = datetime_object.strftime('%Y-%m-%d %H:%M:%S')

                    solution_dict = {
                        "Solution": i+1,
                        "Beschrijving": desc_nl_value,
                        "Datetime": formatted_datetime
                    }
                    solutions_list.append(solution_dict)

                return "success", solutions_list
            else:
                return "no_solutions", []
            
    def create_settings_node(self, pressure, speed, batch_stage, dt, comments, rating, rating_count, approval, sku):
        payload = {
            "statements": [
                {
                    "statement": """
                        MATCH (s:Sku {sku: $sku})
                        CREATE (settings:Settings {pressure: $pressure, speed: $speed, batch_stage: $batch_stage, datetime: $datetime, comments: $comments, rating: $rating, rating_count: $rating_count, approval: $approval})-[:SETTINGS_FOR]->(s)
                        RETURN settings
                    """,
                    "parameters": {
                        "pressure": pressure,
                        "speed": speed,
                        "batch_stage": batch_stage,
                        "datetime": dt,
                        "comments": comments,
                        "rating": rating,
                        "rating_count": rating_count,
                        "approval": approval,
                        "sku": sku
                    },
                }
            ]
        }
        try:
            response = requests.post(self.url, headers=self.headers, data=json.dumps(payload))
            result = response.json()
            if result.get("errors"):
                return "error", [str(e) for e in result.get("errors")]
            
            data = result["results"][0]["data"]
            if len(data) > 0:
                return "success", "Settings node created successfully."
            else:
                return "error", "Failed to create settings node. SKU might not exist."
        except requests.exceptions.RequestException as e:
            return "error", [str(e)]


    def get_settings_by_sku(self, sku):
        payload = {
            "statements": [
                {
                    "statement": """
                        MATCH (settings:Settings)-[:SETTINGS_FOR]->(s:Sku {sku: $sku})
                        RETURN settings
                        ORDER BY settings.datetime DESC
                    """,
                    "parameters": {
                        "sku": sku
                    },
                }
            ]
        }
        try:
            response = requests.post(self.url, headers=self.headers, data=json.dumps(payload))
            result = response.json()
            if result.get("errors"):
                return "error", [str(e) for e in result.get("errors")]

            data = result["results"][0]["data"]
            settings_list = []
            for entry in data:
                row = entry["row"][0]
                unix_timestamp = row["datetime"]
                dt = datetime.fromtimestamp(unix_timestamp)

                # Convert the UTC datetime to CET (Central European Time)
                cet = pytz.timezone("Europe/Paris")
                cet_time = dt.astimezone(cet)

                # Format the datetime object as a human-readable string
                formatted_cet_time = cet_time.strftime("%d-%m-%Y %H:%M:%S")

                settings_dict = {
                    "pressure": row["pressure"],
                    "speed": row["speed"],
                    "batch_stage": row["batch_stage"],
                    "datetime": formatted_cet_time,
                    "comments": row["comments"],
                    "rating": row["rating"],
                    "rating_count": row["rating_count"],
                    "approval": row["approval"]
                }
                settings_list.append(settings_dict)

            if len(settings_list) == 0:
                return None, "No settings found for the given SKU."
            elif len(settings_list) == 1:
                return "success", [settings_list[0]]
            else:
                return "success", settings_list
        except requests.exceptions.RequestException as e:
            return "error", [str(e)]

